""" Utilities for manipulating asccidoc (asciidoctor) documents

Typical DocTestParser Expression objects:
{'source': 'import spacy\n',
 'want': '',
 'lineno': 64,
 'indent': 0,
 'options': {},
 'exc_msg': None},
{'source': 'nlp = spacy.load("en_core_web_sm")\n',
 'want': '',
 'lineno': 65,
 'indent': 0,
 'options': {},
 'exc_msg': None},
{'source': 'sentence = \"\"\"The faster Harry got to the store, the faster Harry,\n    the ...',
 'want': '',
 'lineno': 67,
 'indent': 0,
 'options': {},
 'exc_msg': None}
"""
from doctest import DocTestParser
import logging
from pathlib import Path
import re
import nbformat as nbf
from tqdm import tqdm

from nlpia2.text_processing.extractors import parse_args
from nbformat.v4 import new_notebook, new_markdown_cell, new_code_cell, new_output
import sys

log = logging.getLogger(__name__)
HEADER_TEXT = f"""\
# Imports and Settings

>>> import pandas as pd
>>> pd.options.display.max_columns = 3000
"""
HEADER_BLOCKS = [
    dict(source='### Imports and Settings', typ='markdown'),
    dict(source='''\
        >>> import pandas as pd
        >>> pd.options.display.max_columns = 3000
        ''', typ='python')
]

TEST_TEXT = f"""\
# A dataframe

>>> import pandas as pd
>>> pd.options.display.max_columns = 3000
>>> pd.DataFrame([[1,2],[3,4]])
   0  1
0  1  2
1  2  3
"""
TEST_BLOCKS = [
    dict(source='### A dataframe', typ='markdown'),
    dict(source='''\
        >>> pd.DataFrame([[1,2],[3,4]])
        ''', typ='python',
         want='''\
           0  1
        0  1  2
        1  2  3
        ''')
]

# use newlines and groups instead of list of patterns
CODEHEADER = [
    r'^([.][-\w\s\(\)\[\]\|\*\'"!@#{}&^%$+=<,>.?/]+\s*)?\s*$',  # '.Problem setup'
    r'^\[source,\s*python\]\s*$',
    r'^----[-]*\s*$',
]

# Need multiline pattern to capture multiple annotations and blank line after annotations
CODEFOOTER = r'\n(----[-]*\s*\n(<\d+>\s?[^\n]+\n)*)'


def re_matches(patterns, lines, newline='\n', match_objects=True):
    r""" Match multiple regular expressions to consecutive lines of text

    >>> patterns = '^([.][\\w\\s]+)?\s*$\n^\[source\]$\n^[-]+$'.split('\n')
    >>> lines = '----\n.hello\n\n[source]\n----\nworld\n----'.split('\n')
    >>> re_matches(patterns, lines)
    []
    >>> re_matches(patterns, lines[2:])
    [<re.Match object; span=(0, 0), match=''>,
     <re.Match object; span=(0, 8), match='[source]'>,
     <re.Match object; span=(0, 4), match='----'>]
    >>> patterns = r'\n(----[-]*\s*\n((<\d+>\s?[^\n]+)?\n)+)'
    >>> lines = '\n----\n.hello\n\n[source]\n----\nworld\n----\n<1> annot.\n\n'
    >>> re_matches(patterns, lines)

    """
    if isinstance(lines, str):
        lines = lines.splitlines()
    if isinstance(patterns, str):
        it = re.finditer(patterns, newline.join(lines))
        try:
            return next(it).group().splitlines()  # footers can't return a list of match objects
        except StopIteration:
            return []
    matches = []
    for pat, text in zip(patterns, lines):
        match = re.match(pat, text)
        if not match:
            return matches
        matches.append(match)
    if not match_objects:
        return [m.group() for m in matches]
    return matches


def get_examples(text, join_consecutive=True):
    r""" Extract all doctest code and output examples from asciidoc (adoc) text

    >>> text = ">>> import hello\n>>> hello.world()\n'hi'\n"
    >>> get_examples(text)[0]['source']
    'import hello\nhello.world()'
    >>> get_examples(text)[0]['want']
    "'hi'\n"
    """
    dtparser = DocTestParser()
    try:
        examples = dtparser.get_examples(text)
    except ValueError:
        msg = f'Error processing doctests in {text[:40]}...{text[-40:]}'
        log.error(msg)
        raise
    examples = [vars(obj) for obj in examples]
    if not join_consecutive:
        return examples

    blocks = [{}]
    for ex in examples:
        if not blocks[-1]:
            blocks[-1] = ex.copy()
        else:
            blocks[-1]['source'] += ex['source']
            blocks[-1]['want'] += ex['want']
        if blocks[-1]['want'].strip():
            blocks.append({})
    return [d for d in blocks if len(d)]


def rstrip_one_newline(text, newlines='\r\n'):
    """Like text.rstrip('\n').rstrip('\r') but only strips a maximum of 1 character with each strip """
    for c in reversed(newlines):
        if text.endswith(c):
            text = text[:-1]
    return text


def get_code_blocks(text,
                    header_patterns=CODEHEADER,
                    footer_patterns=CODEFOOTER,
                    min_header_len=2,
                    min_footer_len=2,
                    max_block_lines=64):
    if footer_patterns is None:
        footer_patterns = [header_patterns[-1]]
    lines = text.splitlines()
    blocks = []
    i = len(header_patterns)
    while i < len(lines) - 1:
        header_matches = re_matches(header_patterns, lines[i:])
        # some number of header lines must match (2 for adoc, [source,python]\n----\n)
        if not len(header_matches) >= min_header_len:
            i += 1
            continue
        line_number = i
        if header_matches[0].group().strip():
            line_number -= 1
        i += len(header_matches)
        # last line of header must match
        if not re.match(header_patterns[-1], header_matches[-1].group()):
            i += 1
            continue
        # TODO: include docutils examples(parsed code and output examples)
        source_lines = []
        block = {}
        while i < len(lines) - min_footer_len and len(source_lines) < max_block_lines:
            source_lines.append(lines[i])
            i += 1
            footer_matches = re_matches(footer_patterns, lines[i:])
            footer_matches = [getattr(m, 'group', str(m).__str__)() for m in footer_matches]
            if len(footer_matches) >= 1:
                i += len(footer_matches)
                break
            block['footer'] = footer_matches
        block.update(dict(
            preceding_text=lines[line_number - 1],
            preceding_blank_line=lines[line_number],
            line_number=line_number,
            header='\n'.join([
                getattr(m, 'group', str(m).__str__)().rstrip()
                for m in header_matches]),
            prompted_source='\n'.join([
                s.rstrip() for s in source_lines]) + '\n',
            following_blank_line=lines[i],
        ))
        if i + 1 < len(lines):
            block['following_text'] = lines[i + 1]
        examples = get_examples(block['prompted_source'], join_consecutive=True)

        if examples:
            print(len(examples), 0, vars(examples[0]))
            print(len(examples), -1, vars(examples[-1]))
            for example in examples:
                newblock = block.copy()
                # FIXME: also update line_number from examples
                newblock.update(dict(source=example['source'], want=example['want']))
                blocks.append(newblock)
        else:
            block['source'] = block['prompted_source']
            block['want'] = ''
            blocks.append(block)
    return blocks


def adoc_doctests2ipynb(adocs=Path('.'), dest_filepath=None, **kwargs):
    adocs = Path(adocs)
    text = kwargs.pop('text', None) or ''
    if adocs.is_file():
        text = text + '\n' + adocs.read()
    dest_filepath = dest_filepath if not dest_filepath else Path(dest_filepath)
    examples = get_examples(text)

    nb = new_notebook()
    cells = []
    cells.append(new_markdown_cell(f"#### {adocs}"))

    for exp in examples:
        # need to run the doctest parser on a lot of text to get attr names right
        if isinstance(exp, str):
            cells.append(new_markdown_cell(exp))
        if 'text' in exp:
            cells.append(new_markdown_cell(exp['text']))
        if 'prompted_source' in exp:
            new_code_cell(exp['prompted_source'])

    nb['cells'] = cells
    if dest_filepath:
        with dest_filepath.open('w') as f:
            nbf.write(nb, f)
    return nb


def find_title(text, pattern=r'^\s*[=#]\s?(.+)$'):
    r""" First first line that matches pattern (r'^\s*[=#]\s?.+$')"""
    for line in text.splitlines():
        if re.match(pattern, line):
            return line
    for line in text.splitlines():
        if line.strip():
            return line


def adoc2ipynb(filepath=None, dest_filepath=None, text=None, prompt=False, output=False):
    try:
        text = Path(filepath).open().read()
    except (TypeError, OSError, IOError, FileNotFoundError) as e:
        log.error(f'Invalid filepath: {filepath}\n  ERROR: {e}')

    dest_filepath = None if not dest_filepath else Path(dest_filepath)
    blocks = get_code_blocks(text)

    nb = new_notebook()
    cells = []
    title = find_title(text)
    if filepath:
        title = f'[`{filepath.with_suffix("").name}`]({filepath})'
    if filepath:
        cells.append(new_markdown_cell(f"#### {title}"))

    print(len(blocks), filepath)
    print(dest_filepath)
    for block in blocks:
        # need to run the doctest parser on a lot of text to get attr names right
        if len(block['header'].splitlines()) == 3:
            cells.append(new_markdown_cell('#### ' + block['header'].splitlines()[0]))
        if prompt:
            cells.append(new_code_cell(block['prompted_source']))
        else:
            cells.append(new_code_cell(block['source']))
        if output:
            cells.append(new_output(block['output']))

    nb['cells'] = cells
    if dest_filepath:
        with dest_filepath.open('w') as f:
            nbf.write(nb, f)
    return nb


def adocs2notebooks(adoc_dir=Path('.'), dest_dir=None, glob='Chapter-*.adoc'):
    """ Convert a directory of adoc files into jupyter notebooks """
    notebooks = []
    adoc_dir = Path(adoc_dir)
    if not dest_dir:
        dest_dir = adoc_dir.parent / 'notebooks'
    dest_dir = Path(dest_dir)
    dest_dir.mkdir(exist_ok=True, parents=True)
    for adoc_filepath in tqdm(list(adoc_dir.glob(glob))):
        dest_filepath = dest_dir / adoc_filepath.with_suffix('.ipynb').name
        notebooks.append(adoc2ipynb(filepath=adoc_filepath, dest_filepath=dest_filepath))
    return notebooks


def convert(format='ipynb', **kwargs):
    filepath = kwargs.pop('adocs')
    if filepath:
        print(filepath)
        text = Path(filepath).open().read()
        print(len(text))
    else:
        text = TEST_TEXT
    text = HEADER_TEXT + '\n\n' + text
    return dict(nb=adoc2ipynb(text=text), text=text, filepath=filepath)


if __name__ == '__main__':
    if len(sys.argv) > 1:
        kwargs = parse_args(
            format='ipynb',
            description='Convert adoc code blocks and preceding heading text to a jupyter notebook',
            adocs_help='File path to input adoc file',
            output_help='File path to output ipynb file')
        # format = kwargs.pop('format')
        results = convert(**kwargs)
        # print(results)
